import ast
from collections.abc import Mapping
from importlib import import_module
from json import dumps, loads
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union
from uuid import uuid4

from IPython.display import DisplayObject
from nbformat import NotebookNode
from nbformat.v4 import new_code_cell, new_markdown_cell
from pydantic import BaseModel, Field, PrivateAttr, SerializeAsAny, field_validator
from pydantic._internal._model_construction import ModelMetaclass
from strenum import StrEnum

if TYPE_CHECKING:
    from nbprint.config import Configuration

__all__ = (
    "Type",
    "Role",
    "BaseModel",
    "_append_or_extend",
)


# https://github.com/pydantic/pydantic/issues/6423#issuecomment-1967475432
class _SerializeAsAnyMeta(ModelMetaclass):
    def __new__(cls, name: str, bases: tuple[type], namespaces: dict[str, Any], **kwargs) -> ModelMetaclass:
        annotations: dict = namespaces.get("__annotations__", {}).copy()
        for field, annotation in annotations.items():
            if not field.startswith("__"):
                annotations[field] = SerializeAsAny[annotation]
        namespaces["__annotations__"] = annotations
        return super().__new__(cls, name, bases, namespaces, **kwargs)


class Type(BaseModel):
    module: str
    name: str

    @classmethod
    def from_string(cls, st: str) -> "Type":
        module, name = st.rsplit(".", 1)
        return Type(module=module, name=name)

    def to_string(self) -> str:
        return f"{self.module}:{self.name}"

    def type(self) -> type["Type"]:
        return getattr(import_module(self.module), self.name)

    def load(self, **kwargs) -> "Type":
        return self.type()(**kwargs)


class Role(StrEnum):
    UNDEFINED = "undefined"
    CONFIGURATION = "configuration"
    CONTEXT = "context"
    OUTPUTS = "outputs"
    PARAMETERS = "parameters"
    CONTENT = "content"
    PAGE = "page"
    LAYOUT = "layout"


class BaseModel(BaseModel, metaclass=_SerializeAsAnyMeta):
    # type info
    type: Type = Field(alias="_target_")

    # basic metadata
    tags: list[str] = Field(default_factory=list)
    role: Role = Role.UNDEFINED
    ignore: bool = False

    # frontend code
    # This is designed to match anywidget
    css: Optional[Union[str, Path]] = Field(default="")
    esm: Optional[Union[str, Path]] = Field(default="")
    classname: Optional[Union[str, list[str]]] = Field(default="")
    attrs: Optional[Mapping[str, str]] = Field(default_factory=dict)

    # internals
    # Variable to use inside notebook for this model
    _nb_var_name: Optional[str] = PrivateAttr(default="")

    # id to use to reconstitute dom during page building
    _id: str = PrivateAttr(default_factory=lambda: str(uuid4()).replace("-", ""))

    class Config:
        """Pydantic configuration object."""

        arbitrary_types_allowed: bool = False
        extra: str = "ignore"
        validate_assignment: bool = True

    def __init__(self, **kwargs) -> None:
        if "_target_" not in kwargs:
            kwargs["_target_"] = Type(module=self.__class__.__module__, name=self.__class__.__name__)
        super().__init__(**kwargs)

    @field_validator("css", mode="before")
    @classmethod
    def convert_css_string_or_path_to_string_or_path(cls, v) -> str:
        if isinstance(v, str) and v.strip().endswith(".css"):
            # TODO: resolve relative to class?
            v = Path(v).resolve().read_text()
        return v

    @field_validator("esm", mode="before")
    @classmethod
    def convert_esm_string_or_path_to_string_or_path(cls, v) -> str:
        if isinstance(v, str):
            v = v.strip()
            if v.endswith((".js", ".mjs")):
                # TODO: resolve relative to class?
                v = Path(v).resolve().read_text()
        return v

    @field_validator("type", mode="before")
    @classmethod
    def convert_type_string_to_module_and_name(cls, v) -> Type:
        if isinstance(v, str):
            return Type.from_string(v)
        return v

    @property
    def nb_var_name(self) -> str:
        if self._nb_var_name:
            return self._nb_var_name
        name = self.__class__.__name__.lower()
        if name.startswith("nbprint"):
            return name.replace("nbprint", "nbprint_")
        return f"nbprint_{name}"

    def _recalculate_nb_var_name(self, nb_vars: set) -> None:
        attempt = 0
        test_nb_var_name = self.nb_var_name
        while test_nb_var_name in nb_vars:
            test_nb_var_name = f"{self.nb_var_name}{attempt}"
            attempt += 1
        # reset var name
        self._nb_var_name = test_nb_var_name
        # put in set
        nb_vars.add(self._nb_var_name)

    @staticmethod
    def _to_type(value, model_type=None) -> BaseModel:
        if value is None:
            value = {}

        if model_type is None and "_target_" in value:
            # derive type from instantiation
            model_type = BaseModel

        if isinstance(value, dict):
            return model_type(**value).type.load(**value)

        return value

    @classmethod
    def from_json(cls, json) -> BaseModel:
        data = loads(json)
        return cls(**data)

    def __call__(self, **_) -> Optional[DisplayObject]:
        """Execute this model inside of a notebook

        Args:
        ----
            ctx (_type_, optional): _description_. Defaults to None.

        """
        return self

    def render(self, **_) -> None:
        """Called during notebook generation only, this should run any necessary post-processing. Pre-processing should go in __init__"""

    def generate(
        self,
        metadata: dict,
        config: Optional["Configuration"],
        parent: Optional["BaseModel"] = None,
        attr: str = "",
        counter: Optional[int] = None,
        **_,
    ) -> Optional[Union[NotebookNode, list[NotebookNode]]]:
        """Generate a notebook node for this model.
        This will be called before the runtime of the notebook, use it for code generation.

        Args:
        ----
            metadata (dict): common cell metadata

        Returns:
        -------
            NotebookNode: the content of the notebook node

        """
        return self._base_generate(metadata=metadata, config=config, parent=parent, attr=attr, counter=counter)

    def _base_set_nbprint_metadata(self, cell: NotebookNode) -> None:
        cell.metadata.nbprint.id = self._id
        cell.metadata.nbprint.role = self.role or "undefined"
        cell.metadata.nbprint.type = self.type.to_string()
        cell.metadata.nbprint.ignore = self.ignore or False
        if cell.metadata.nbprint.ignore and self.role in (Role.PARAMETERS,):
            # Don't collapse
            ...
        elif cell.metadata.nbprint.ignore or self.role in (
            Role.CONFIGURATION,
            Role.CONTEXT,
            Role.PAGE,
            Role.UNDEFINED,
        ):
            # Collapse cell by default
            cell.metadata["jupyter"] = {
                "source_hidden": True,
                "outputs_hidden": True,
            }
            cell.metadata.collapsed = True
        cell.metadata.nbprint.css = self.css or ""
        cell.metadata.nbprint.esm = self.esm or ""
        cell.metadata.nbprint.class_selector = f'{cell.metadata.nbprint.type.replace(":", "-").replace(".", "-")}'
        cell.metadata.nbprint.element_selector = f"{cell.metadata.nbprint.class_selector}-{self._id}"
        cell.metadata.nbprint["class"] = f"nbprint {cell.metadata.nbprint.class_selector} {cell.metadata.nbprint.element_selector} " + (
            " ".join(self.classname) if isinstance(self.classname, list) else self.classname or ""
        )
        cell.metadata.nbprint.attrs = " ".join(f"{k}={dumps(v)}" for k, v in (self.attrs or {}).items())

    def _base_generate_meta(self, metadata: Optional[dict] = None) -> Optional[NotebookNode]:
        cell = new_code_cell(metadata=metadata)
        cell.metadata.tags = list(set(["nbprint"] + (self.tags or [])))

        # TODO: consolidate with self.model_dump_json(by_alias=True)?
        self._base_set_nbprint_metadata(cell)
        cell.metadata.nbprint.data = self.model_dump_json(by_alias=True)
        return cell

    def _base_generate_md_meta(self, metadata: Optional[dict] = None) -> Optional[NotebookNode]:
        cell = new_markdown_cell(metadata=metadata)
        cell.metadata.tags = list(set(["nbprint"] + (self.tags or [])))
        self._base_set_nbprint_metadata(cell)
        return cell

    def _base_generate(
        self,
        metadata: dict,
        config: "Configuration",
        parent: Optional["BaseModel"] = None,
        attr: str = "",
        counter: Optional[int] = None,
    ) -> Optional[NotebookNode]:
        from nbprint.config import Configuration

        # trigger any pre-cell generation logic
        self.render(config=config)

        cell = self._base_generate_meta(metadata=metadata)
        mod = ast.Module(body=[], type_ignores=[])

        assert config is not None
        self._recalculate_nb_var_name(config._nb_vars)

        if parent and parent.ignore is False:
            # TODO: should this go in a standard location?
            # set in metadata
            cell.metadata.nbprint["parent-id"] = parent._id

        if parent and attr:
            # now construct accessor
            value = ast.Attribute(value=ast.Name(id=parent.nb_var_name, ctx=ast.Load()), attr=attr, ctx=ast.Load())

            if counter is not None:
                value = ast.Subscript(
                    value=value,
                    slice=ast.Constant(value=counter),
                )
            mod.body.append(
                ast.Assign(
                    targets=[ast.Name(id=self.nb_var_name, ctx=ast.Store())],
                    value=value,
                    lineno=1,
                )
            )
        else:
            data = self.model_dump_json(by_alias=True)
            mod.body.append(
                ast.ImportFrom(
                    module=self.type.module,
                    names=[ast.alias(name=self.type.name)],
                    level=0,
                )
            )

            mod.body.append(
                ast.Assign(
                    targets=[ast.Name(id=self.nb_var_name, ctx=ast.Store())],
                    value=ast.Call(
                        func=ast.Attribute(value=ast.Name(id=self.type.name, ctx=ast.Load()), attr="from_json", ctx=ast.Load()),
                        args=[ast.parse(dumps(data), mode="eval").body],
                        keywords=[],
                    ),
                    lineno=2,
                )
            )

        # Determine whether to pass context into the call.
        # If we have a config object, and that config object has a context, and we've previously made the context object,
        # and we ourselves are not a config, then pass in
        call_with_context = (
            config.context.nb_var_name
            if config and config.context and config.context._context_generated and not isinstance(self, Configuration)
            else "None"
        )

        mod.body.append(
            ast.Expr(
                value=ast.Call(
                    func=ast.Name(id=self.nb_var_name, ctx=ast.Load()),
                    args=[],
                    keywords=[ast.keyword(arg="ctx", value=ast.Name(id=call_with_context, ctx=ast.Load()))],
                )
            )
        )
        # this line just puts the object as the last item, lets call it always instead
        # mod.body.append(ast.Expr(value=ast.Name(id=self.nb_var_name, ctx=ast.Load())))

        source = ast.unparse(mod).replace('"', '\\"')
        cell.source = source
        return cell

    def _base_generate_md(self, metadata: dict, config: "Configuration", parent: Optional["BaseModel"] = None) -> Optional[NotebookNode]:
        # trigger any pre-cell generation logic
        self.render(config=config)

        cell = self._base_generate_md_meta(metadata=metadata)
        # TODO: should this go in a standard location?
        # set in metadata
        if parent and parent.ignore is False:
            cell.metadata.nbprint["parent-id"] = parent._id
        return cell

    def __repr__(self) -> str:
        # Truncate the output for now
        return f"<{self.__class__.__name__}>"


def _append_or_extend(cells: list, cell_or_cells: Union[NotebookNode, list[NotebookNode]]) -> None:
    if isinstance(cell_or_cells, list):
        cells.extend(cell_or_cells)
    elif cell_or_cells:
        cells.append(cell_or_cells)
    # None, ignore



from datetime import datetime
from pathlib import Path

from nbprint.cli import run_hydra


def test_outputs():
    config = run_hydra(str(Path(__file__).parent / "files" / "basic.yaml"), dry_run=True)
    assert config.outputs.naming == "{{name}}-{{date}}-{{datetime}}-{{uuid}}-{{sha}}"
    path = config.outputs.run(config=config, gen=config.generate())
    assert len(path.name) == 150
    today = datetime.now()
    assert path.name.startswith(f"basic-{today.year}-{today.month}-{today.day}")
    assert config.outputs._get_sha(config=config) in path.name



import os.path
from typing import Optional

from IPython.display import HTML
from pydantic import FilePath

from nbprint import ContentCover, ContentImage

__all__ = (
    "ExampleNBPrintLogo",
    "ExampleCoverPageContent",
)


class ExampleNBPrintLogo(ContentImage):
    path: FilePath = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "..", "docs", "img", "logo-light.png"))


class ExampleCoverPageContent(ContentCover):
    logo: Optional[ContentImage]
    title: Optional[str] = ""
    subtitle: Optional[str] = ""

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"""
            {self.logo()._repr_html_()}
            <h1>{self.title}</h1>
            <h2>{self.subtitle}</h2>
            <h3>{ctx.string}</h3>
            <p class="pagebreak"></p>
        """)



from typing import Literal, Union

from IPython.display import HTML
from pydantic import Field

from nbprint import Content

__all__ = ("SeabornDisplayConfiguration",)


class SeabornDisplayConfiguration(Content):
    style: Literal["white", "dark", "whitegrid", "darkgrid", "ticks"] = Field(default="whitegrid")
    context: Literal["paper", "notebook", "talk", "poster"] = Field(default="notebook")
    palette: Union[str, list[str]] = Field(default="tab10")

    def __call__(self, **_) -> HTML:
        import seaborn as sns

        sns.set_theme(self.context, self.style, self.palette)
        return HTML("")



from typing import Optional

from .common import _DirectionalSize
from .css import _BaseCss

__all__ = (
    "Padding",
    "Margin",
    "Spacing",
)


class Padding(_DirectionalSize):
    def __str__(self) -> str:
        return "\n".join(
            f"padding-{direction}: {getattr(self, direction, '')}px;"
            for direction in ("right", "left", "top", "bottom")
            if getattr(self, direction, "")
        ).strip()


class Margin(_DirectionalSize):
    def __str__(self) -> str:
        return "\n".join(
            f"margin-{direction}: {getattr(self, direction, '')}px;"
            for direction in ("right", "left", "top", "bottom")
            if getattr(self, direction, "")
        ).strip()


class Spacing(_BaseCss):
    padding: Optional[Padding] = None
    margin: Optional[Margin] = None

    def __str__(self) -> str:
        ret = ""
        if self.padding:
            ret += f"{self.padding}\n"
        if self.margin:
            ret += f"{self.margin}\n"
        return ret.strip()



from IPython.display import HTML
from pydantic import Field

from nbprint.config.base import Role

from .base import Content

__all__ = (
    "ContentLayout",
    "ContentInlineLayout",
    "ContentFlexColumnLayout",
    "ContentFlexRowLayout",
)


class _ContentFlexLayout(Content):
    # component to split into certain number of columns
    # count: int = 1
    sizes: list[float] = Field(default=[])

    # override role
    role: Role = Role.LAYOUT

    css: str = ":scope { display: flex; }"
    esm: str = """
function render(meta, elem) {
    let data = JSON.parse(meta.data);

    if (data.sizes.length <= 0 )
        return;

    let size_index = 0;

    Array.from(elem.children).forEach((child) => {
        let size = data.sizes[size_index];

        if (!size)
            return;

        // TODO:these are hacks to determine
        // if it should be included
        let output_children = (child.querySelector(".jp-OutputArea-output") || {}).children || [];
        if (Array.from(child.classList).includes("nbprint")) {
            child.style.flex = `${size}`;
            size_index += 1;
        } else if (Array.from(output_children).length > 0) {
            child.style.flex = `${size}`;
            size_index += 1;
        } else {
            child.style.flex = "0";
        }
    });
}
"""
    attrs: dict = Field(default_factory=dict)

    def __call__(self, **_) -> HTML:
        # return empty html just for placeholder
        return HTML("")


class ContentLayout(Content):
    # override role
    role: Role = Role.LAYOUT

    def __call__(self, **_) -> HTML:
        # return empty html just for placeholder
        return HTML("")


class ContentInlineLayout(Content):
    # override role
    role: Role = Role.LAYOUT

    css: str = ":scope { display: block; }"
    esm: str = """
function render(meta, elem) {
    let data = JSON.parse(meta.data);

    Array.from(elem.children).forEach((child) => {
        // TODO:these are hacks to determine
        // if it should be included
        let output_children = (child.querySelector(".jp-OutputArea-output") || {}).children || [];
        if (Array.from(child.classList).includes("nbprint")) {
            child.style.display = "inline-block";
            child.style.float = "left";
        } else if (Array.from(output_children).length > 0) {
            child.style.display = "inline-block";
            child.style.float = "left";
        }
    });
}
"""
    attrs: dict = Field(default_factory=dict)

    def __call__(self, **_) -> HTML:
        # return empty html just for placeholder
        return HTML("")


class ContentFlexColumnLayout(_ContentFlexLayout):
    css: str = """
:scope { display: flex; flex-direction: column; break-inside: auto; }
"""


class ContentFlexRowLayout(_ContentFlexLayout):
    css: str = """
:scope { display: flex; flex-direction: row; break-inside: auto; }
"""



from typing import Optional

from IPython.display import HTML
from pydantic import Field

from nbprint import Content

__all__ = ("PandasDisplayConfiguration",)


class PandasDisplayConfiguration(Content):
    max_columns: Optional[int] = Field(default=None)
    max_rows: int = Field(default=100)

    def __call__(self, **_) -> HTML:
        import pandas as pd

        pd.set_option("display.max_columns", self.max_columns)
        pd.set_option("display.max_rows", self.max_rows)
        return HTML("")



from IPython.display import HTML

from nbprint import Content

__all__ = (
    "ExampleFinanceAuthor",
    "ExampleFinanceStockEarningsTable",
    "ExampleFinanceStockHeadline",
    "ExampleFinanceStockPTUpdate",
    "ExampleFinanceStockQuickStats",
    "ExampleFinanceStockSubHeadline",
    "ExampleFinanceStockWhatsChanged",
    "ExampleFinanceTitle",
    "ExampleFinanceStockBody",
    "ExampleFinanceFirstPageDisclosures",
    "ExampleFinanceDisclosures",
)


class ExampleFinanceTitle(Content):
    title: str = ""
    subtitle: str = ""
    date: str = ""
    css: str = """

h1 {
  font-family: sans-serif;
  font-size: 25px !important;
  padding-right: 15px;
  border-right: 2px solid black;
}

h2 {
  font-family: sans-serif;
  text-transform: uppercase;
  font-weight: bold !important;
  font-size: 12px !important;
  padding-left: 15px;
}

span {
  color: grey;
  font-size: 15px !important;
  font-style: italic;
  padding-top: 10px;
  padding-bottom: 10px;
}
    """

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"""
        <div class="row">
            <h1>{self.title}</h1>
            <h2>{self.subtitle}</h2>
        </div>
        <div class="row">
            <span>{self.date}</span>
        </div>
        """)


class ExampleFinanceStockHeadline(Content):
    company_name: str = ""
    region: str = ""
    byline: str = ""
    color: str = ""
    css: str = """
div.row {
  padding-bottom: 10px;
  margin-top: 12px !important;
  margin-bottom: 12px !important;
}

h2:nth-child(1) {
  padding-right: 15px;

}

h2:nth-child(2) {
  padding-left: 15px;
}

h1 {
  font-weight: lighter !important;
  margin-bottom: 12px !important;
}
"""

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"""
        <div class="row">
            <h2 style="color: {self.color}; border-right: 2px solid {self.color}">{self.company_name}</h2>
            <h2 style="color: {self.color};">{self.region}</h2>
        </div>
        <h1>{self.byline}</h1>
        """)


class ExampleFinanceAuthor(Content):
    name: str = ""
    title: str = ""
    email: str = ""
    phone: str = ""
    css: str = """

div.column > span:nth-child(1) {
  font-size: 12px;
  font-weight: bold;
}

div.column > span:nth-child(2) {
  font-size: 10px;
  text-transform: uppercase;
}

div.row {
  justify-content: space-between;
}

div.row > span:nth-child(1) {
  font-size: 10px;
}

div.row > span:nth-child(2) {
  font-family: monospace;
  font-size: 10px;
}

    """

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"""
            <div class="column">
                <span>{self.name}</span>
                <span>{self.title}</span>
                <div class="row">
                    <span>{self.email}</span>
                    <span>{self.phone}</span>
                </div>
            </div>
        """)


class ExampleFinanceStockQuickStats(Content):
    ticker: str = ""
    company_name: str = ""
    color: str = ""

    sector: str = ""
    country: str = ""
    region: str = ""

    rating: str
    view: str
    price_target: float

    share_price: float
    market_cap: str
    range_min: float
    range_max: float

    css: str = """
div.row {
  justify-content: space-between;
  font-size: 10px;
}

div.column > div.column > div.row:nth-child(1),
div.column > div.column > div.row:nth-child(2),
div.column > div.column > div.row:nth-child(3) {
  font-weight: bold;
}
"""

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"""
            <div class="column pt20">
                <div style="padding-left: 5px; background-color: {self.color}; color: white;">
                    <span>{self.company_name} ( {self.ticker} )</span>
                </div>
                <div class="row" style="font-size: 9px; justify-content: unset;">
                    <span>{self.sector}</span>
                    <span>&nbsp;/&nbsp;</span>
                    <span style="font-weight: bold;">{self.country}</span>
                </div>
                <div class="column">
                    <div class="row">
                        <span>Stock Rating</span>
                        <span>{self.rating}</span>
                    </div>
                    <div class="row">
                        <span>Industry View</span>
                        <span>{self.view}</span>
                    </div>
                    <div class="row">
                        <span>Price target</span>
                        <span>${self.price_target:.2f}</span>
                    </div>
                    <div class="row">
                        <span>Share Price</span>
                        <span>${self.share_price:.0f}</span>
                    </div>
                    <div class="row">
                        <span>Market Cap (mm)</span>
                        <span>${self.market_cap}</span>
                    </div>
                    <div class="row">
                        <span>52-Week Range</span>
                        <span>${self.range_min:.2f} - {self.range_max:.2f}</span>
                    </div>
                </div>
            </div>
        """)


class ExampleFinanceStockEarningsTable(Content):
    ticker: str = ""
    company_name: str = ""

    css: str = """
:scope {
  margin: 0px !important;
  padding: 0px !important;
  break-before: avoid;
}

table {
  font-size: 10px !important;
}
table thead tr {
  font-weight: bold;
}
table tbody tr td:nth-child(1){
  font-weight: bold;
}
"""

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML("""
            <table>
                <thead>
                    <tr>
                        <td>Fiscal Year Ending</td>
                        <td>3/24</td>
                        <td>3/25e</td>
                        <td>3/26e</td>
                        <td>3/27e</td>
                    </tr>
                </thead>
                <tbody>
                    <tr>
                        <td>EPS ($)</td>
                        <td>(0.32)</td>
                        <td>(0.14)</td>
                        <td>0.11</td>
                        <td>0.31</td>
                    </tr>
                    <tr>
                        <td>Prior EPS ($)</td>
                        <td>(0.23)</td>
                        <td>(0.15)</td>
                        <td>0.03</td>
                        <td>0.13</td>
                    </tr>
                    <tr>
                        <td>P/E</td>
                        <td>NM</td>
                        <td>NM</td>
                        <td>NM</td>
                        <td>NM</td>
                    </tr>
                </tbody>
            </table>
        """)


class ExampleFinanceStockPTUpdate(Content):
    ticker: str = ""
    rating: str
    view: str
    price_target: float

    css: str = """
:scope div.row {
  margin-top: 15px;
  margin-bottom: 15px;
  justify-content: space-between;
}

div.row > div:nth-child(2),
div.row > div:nth-child(3) {
  padding-left: 20px;
  border-left: 1px solid black;
}

div.row > div > span:nth-child(1) {
  font-size: 10px;
}

div.row > div > span:nth-child(2) {
  font-weight: lighter !important;
  font-size: 17px;
  padding-left: 5px;
}

div.row div:nth-child(1) span:nth-child(1)::before {
  font-family: "Font Awesome 6 Free";
  font-weight: 900;
  content: '\\f201';
  display: none;
  width: 20px;
}

div.row div:nth-child(2) span:nth-child(1)::before {
  font-family: "Font Awesome 6 Free";
  content: '\\f06e';
  display: none;
}

div.row div:nth-child(3) span:nth-child(1)::before {
  font-family: "Font Awesome 6 Free";
  content: '\\f192';
  display: none;
}

svg {
  width: 15px;
  margin-right: 3px;
}
"""

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"""
        <div class="row">
            <div class="column">
                <span class="stock-rating">Stock Rating</span>
                <span>{self.rating}</span>
            </div>
            <div class="column">
                <span>Industry View</span>
                <span>{self.view}</span>
            </div>
            <div class="column">
                <span>Price Target</span>
                <span>${self.price_target:.2f}</span>
            </div>
        </div>
        """)


class ExampleFinanceStockSubHeadline(Content):
    text: str = ""

    css: str = """
:scope span {
  font-size: 20px;
}
"""

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"""
            <span>{self.text}</span>
        """)


class ExampleFinanceStockWhatsChanged(Content):
    ticker: str = ""
    company_name: str = ""
    color: str = ""
    price_target: float
    price_target_old: float

    css: str = """
:scope {
  margin-top: 10px;
  margin-bottom: 10px;
}

:scope span {
  font-size: 15px;
}

div.row {
  justify-content: space-between;
}

table {
  flex: 1;
  margin-left: 7px !important;
}

thead {
  border-bottom: 1px solid black;
  font-weight: bold;
}

td {
  text-align: left !important;
  padding: 0px !important;
}

div.whatschangedbox {
  text-transform: uppercase;
  font-weight: 100;
  padding: 2px;
}
"""

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"""
          <div class="row">
            <div class="column whatschangedbox" style="padding: 5px; background-color: {self.color}; color: white;">
              <span>What's</span>
              <span>Changed</span>
            </div>
            <table>
              <thead>
                <tr>
                  <td>{self.company_name}</td>
                  <td>From</td>
                  <td>To</td>
                </tr>
                <tr>
                  <td>({self.ticker})</td>
                  <td></td>
                  <td></td>
                </tr>
              </thead>
              <tbody>
                <tr>
                  <td>Price Target</td>
                  <td>{self.price_target_old:.2f}</td>
                  <td style="font-weight: bold;">{self.price_target:.2f}</td>
                </tr>
              </tbody>
            </table>
          </div>
        """)


class ExampleFinanceStockBody(Content):
    our_take: str = ""
    review: str = ""
    outlook: str = ""

    css: str = """
p {
  margin: auto !important;
}
"""

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"""
        <div class="column">
          <span><b>Our take</b>: {self.our_take}</span>
          <span><b>Quarter in review</b>: {self.review}</span>
          <span><b>Outlook</b>: {self.outlook}</span>
        </div>
        """)


class ExampleFinanceFirstPageDisclosures(Content):
    css: str = """

:scope {
  /* margin-top: 350px; */
  line-height: 1 !important;
}

span {
  font-size: 8px;
}

span:nth-child(2) {
  font-weight: bold;
}

"""

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML("""
  <div class="column">
    <span>Jefferies Morgan does and seeks to do business with companies covered in Jefferies Morgan Research. As a result, investors should be aware that the firm may have a conflict of interest that could affect the objectivity of Jefferies Morgan Research. Investors should consider Jefferies Morgan Research as only a single factor in making their investment decision.</span>
    <span>For analyst certification and other important disclosures, refer to the Disclosure Section, located at the end of this report.</span>
  </div>
        """)


class ExampleFinanceDisclosures(Content):
    css: str = """
:scope h1 {
  font-size: 12px;
}

:scope p {
  font-size: 8px;
}
"""

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML("""
                <h1>Important Disclosures</hw1>
<p>
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
</p>
<p>
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
</p>
<p>
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
</p>
        """)



from typing import Optional

from strenum import StrEnum

from .css import _BaseCss

__all__ = (
    "FlexDirection",
    "VerticalAlignment",
    "HorizontalAlignment",
    "Justify",
    "DisplayKind",
    "FlexOptions",
    "Display",
)


class FlexDirection(StrEnum):
    row = "row"
    column = "column"


class VerticalAlignment(StrEnum):
    top = "top"
    center = "center"
    bottom = "bottom"


class HorizontalAlignment(StrEnum):
    left = "left"
    center = "center"
    right = "right"


class Justify(StrEnum):
    normal = "normal"
    center = "center"
    start = "start"
    end = "end"
    left = "left"
    right = "right"
    space_between = "space-between"
    space_around = "space-around"
    space_evenly = "space-evenly"


class DisplayKind(StrEnum):
    none = "none"
    flex = "flex"


class FlexOptions(_BaseCss):
    flex_direction: Optional[FlexDirection]
    justify: Optional[Justify]


class Display(_BaseCss):
    display: Optional[DisplayKind]
    flex_options: Optional[FlexOptions]



from pydantic import Field

from nbprint.config import Parameters

# from datetime import datetime
# from typing import Dict, list, Optional


class ExampleFinanceParameters(Parameters):
    ticker: str
    authors: list[str] = Field(default=[])
    color: str



from IPython.display import DisplayObject, display


class DisplayList(DisplayObject):
    def __init__(self, *objs) -> None:
        self._objs = objs

    def _repr_html_(self) -> str:
        for o in self._objs:
            display(o)
        return ""



from typing import TYPE_CHECKING, Optional, Union

from nbformat import NotebookNode
from pydantic import Field, field_validator

from .base import BaseModel, Role, Type, _append_or_extend
from .common import PageOrientation, PageSize, Style
from .exceptions import NBPrintNullCellError

if TYPE_CHECKING:
    from .core.config import Configuration

__all__ = (
    "Page",
    "PageNumber",
    "PageRegion",
    "PageRegionContent",
)


class PageRegionContent(BaseModel):
    content: Optional[str] = ""

    # common
    tags: list[str] = Field(default=["nbprint:page"])
    role: Role = Role.PAGE
    ignore: bool = True

    def __str__(self) -> str:
        if self.content:
            return f"content: {self.content};"
        return ""


class PageNumber(PageRegionContent):
    content: Optional[str] = "counter(page)"


class PageRegion(BaseModel):
    _region: str = ""
    content: Optional[PageRegionContent] = Field(default_factory=PageNumber)
    style: Optional[Style] = None

    css: str = ""

    # common
    tags: list[str] = Field(default=["nbprint:page"])
    role: Role = Role.PAGE
    ignore: bool = True

    def generate(self, metadata: dict, config: "Configuration", parent: "BaseModel", attr: str, **_) -> NotebookNode:
        cell = super().generate(metadata=metadata, config=config, parent=parent, attr=attr)
        cell.metadata.tags.append(f"nbprint:page:{attr}")
        return cell


class Page(BaseModel):
    top: Optional[PageRegion] = None
    top_left: Optional[PageRegion] = None
    top_right: Optional[PageRegion] = None
    bottom: Optional[PageRegion] = None
    bottom_left: Optional[PageRegion] = None
    bottom_right: Optional[PageRegion] = None
    left: Optional[PageRegion] = None
    left_top: Optional[PageRegion] = None
    left_bottom: Optional[PageRegion] = None
    right: Optional[PageRegion] = None
    right_top: Optional[PageRegion] = None
    right_bottom: Optional[PageRegion] = None

    size: Optional[PageSize] = Field(default=PageSize.letter)
    orientation: Optional[PageOrientation] = Field(default=PageOrientation.portrait)

    pages: Optional[list["Page"]] = Field(default_factory=list)

    css: str = ""

    @classmethod
    def convert_region_from_obj(cls, v, region) -> PageRegion:
        ret = BaseModel._to_type(v, PageRegion)
        ret._region = region
        base_css_scope = "@page {{ @{region} {{ {content} }} }}".format(region=region, content=str(ret.content or ""))
        ret.css += base_css_scope
        return ret

    @field_validator("top", mode="before")
    @classmethod
    def convert_top_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "top-center")

    @field_validator("top_left", mode="before")
    @classmethod
    def convert_top_left_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "top-left")

    @field_validator("top_right", mode="before")
    @classmethod
    def convert_top_right_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "top-right")

    @field_validator("bottom", mode="before")
    @classmethod
    def convert_bottom_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "bottom-center")

    @field_validator("bottom_left", mode="before")
    @classmethod
    def convert_bottom_left_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "bottom-left")

    @field_validator("bottom_right", mode="before")
    @classmethod
    def convert_bottom_right_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "bottom-right")

    @field_validator("left", mode="before")
    @classmethod
    def convert_left_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "left-center")

    @field_validator("left_top", mode="before")
    @classmethod
    def convert_left_top_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "left-top")

    @field_validator("left_bottom", mode="before")
    @classmethod
    def convert_left_bottom_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "left-bottom")

    @field_validator("right", mode="before")
    @classmethod
    def convert_right_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "right-center")

    @field_validator("right_top", mode="before")
    @classmethod
    def convert_right_top_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "right-top")

    @field_validator("right_bottom", mode="before")
    @classmethod
    def convert_right_bottom_from_obj(cls, v) -> PageRegion:
        return Page.convert_region_from_obj(v, "right-bottom")

    def render(self, **_) -> None:
        if "@page { size:" not in self.css:
            self.css = (
                self.css
                + f"""
@page {{ size: {self.size.value} {self.orientation.value}; }}
            """
            )

    def generate(
        self, metadata: dict, config: "Configuration", parent: "BaseModel", attr: str = "page", **_
    ) -> Optional[Union[NotebookNode, list[NotebookNode]]]:
        cells = []

        # parent and config should be equal for the global page layout
        assert parent == config

        main_cell = super()._base_generate(metadata=metadata, config=config, parent=config, attr=attr or "page")
        main_cell.metadata.tags.append("nbprint:page")
        main_cell.metadata.nbprint.role = "page"
        main_cell.metadata.nbprint.ignore = True
        cells.append(main_cell)

        for set_attr in (
            "top",
            "top_left",
            "top_right",
            "bottom",
            "bottom_left",
            "bottom_right",
            "left",
            "left_top",
            "left_bottom",
            "right",
            "right_top",
            "right_bottom",
        ):
            if getattr(self, set_attr) is not None:
                # pass in `self` as `parent here`
                _append_or_extend(cells, getattr(self, set_attr).generate(metadata=metadata, config=config, parent=self, attr=set_attr))
        for i, cell in enumerate(self.pages):
            _append_or_extend(
                cells,
                cell.generate(metadata=metadata, config=config, parent=self, attr="pages", counter=i),
            )
        for cell in cells:
            if cell is None:
                raise NBPrintNullCellError
        return cells

    @field_validator("pages", mode="before")
    @classmethod
    def convert_pages_from_obj(cls, v) -> "Page":
        if v is None:
            return []
        if isinstance(v, list):
            for i, element in enumerate(v):
                if isinstance(element, str):
                    v[i] = Page(type=Type.from_string(element))
                elif isinstance(element, dict):
                    v[i] = BaseModel._to_type(element)
        return v



from datetime import date, datetime
from typing import Optional

from pydantic import Field

from nbprint.config import Parameters

__all__ = ("ExampleParameters", "ExampleParametersVariousTypes")


class ExampleParameters(Parameters):
    string: str
    a: Optional[int] = 4
    b: Optional[float] = 4.5
    c: list[int] = Field(default=[1, 2, 3])
    d: dict[str, int] = Field(default={"A": 1, "B": 2, "C": 3})


class ExampleParametersVariousTypes(Parameters):
    a: bool
    b: str
    c: int
    d: float
    e: date
    f: datetime
    g: ExampleParameters



from io import BytesIO

import dominate.tags as dt
import dominate.util as du
from IPython.display import HTML
from nbconvert.filters.pandoc import convert_pandoc

from .image import Image


def _html(text, color="") -> HTML:
    """Print in html"""
    text = convert_pandoc(text, "markdown+tex_math_double_backslash", "html")

    if color:
        d = dt.div()
        d.attributes["style"] = f"color: {color};"
        d.appendChild(du.raw(text))
    else:
        d = du.raw(text)

    return HTML(str(d))


def hr() -> HTML:
    """Horizontal rule"""
    return HTML(str(dt.hr()))


def newpage() -> HTML:
    """Make a new page. in html, this just does a horizontal rule"""
    p = dt.p()
    p.attributes["style"] = "page-break-before: always;"
    return _html(str(p))


def _make(text, h_type, **kwargs) -> dt.html_tag:
    h = getattr(dt, h_type)(text)
    h.attributes.update(**kwargs)
    return h


def p(text, **kwargs) -> HTML:
    return HTML(str(_make(text, "p", **kwargs)))


def h1(text, **kwargs) -> HTML:
    return HTML(str(_make(text, "h1", **kwargs)))


def h2(text, **kwargs) -> HTML:
    return HTML(str(_make(text, "h2", **kwargs)))


def h3(text, **kwargs) -> HTML:
    return HTML(str(_make(text, "h3", **kwargs)))


def h4(text, **kwargs) -> HTML:
    return HTML(str(_make(text, "h4", **kwargs)))


def h5(text, **kwargs) -> HTML:
    return HTML(str(_make(text, "h5", **kwargs)))


def plot(fig) -> Image:
    imgdata = BytesIO()
    fig.savefig(imgdata)
    imgdata.seek(0)
    return Image(imgdata.read())



import ast
from json import dumps

from nbformat import NotebookNode
from pydantic import Field

from nbprint.config.base import BaseModel, Role

__all__ = ("Parameters",)


class Parameters(BaseModel):
    tags: list[str] = Field(default=["parameters", "nbprint:parameters"])
    role: Role = Role.PARAMETERS
    ignore: bool = True

    def generate(self, metadata: dict, **_) -> NotebookNode:
        cell = self._base_generate_meta(metadata=metadata)
        # if nb_vars:
        #     # add parameter variable
        #     nb_vars.add(k)

        mod = ast.Module(body=[], type_ignores=[])

        # NOTE: use model_dump(mode="json") here to be compatible with
        # papermill-based json parameters
        for i, (k, v) in enumerate(self.model_dump(mode="json").items()):
            if k in ("type", "tags", "role", "ignore"):
                continue
            if isinstance(v, bool):
                # Handle separately
                to_write = str(v)
            else:
                to_write = dumps(v)
                if isinstance(to_write, str) and (": true," in to_write or ": false," in to_write):
                    to_write = to_write.replace(": true,", ": True,").replace(": false,", ": False,")
            mod.body.append(
                ast.Assign(
                    targets=[ast.Name(id=k, ctx=ast.Store())],
                    value=ast.parse(to_write, mode="eval").body,
                    lineno=i,
                )
            )
        source = ast.unparse(mod).replace('"', '\\"')
        cell.source = source
        return cell



from typing import Optional

import pandas as pd

from nbprint import Context

__all__ = ("ExampleContext",)


class ExampleContext(Context):
    string: str = ""
    df: Optional[pd.DataFrame] = None



import os.path

import pytest

from nbprint import Configuration
from nbprint.cli import run_hydra


def _example_folder_does_not_exist():
    return not os.path.exists("examples")


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples not present - skipping examples tests")
def test_basic_e2e():
    config = Configuration.load("examples/basic.yaml", "basic")
    config.run()


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples not present - skipping examples tests")
def test_inline_e2e():
    config = Configuration.load("examples/inline.yaml", "inline")
    config.run()


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples not present - skipping examples tests")
def test_landscape_e2e():
    config = Configuration.load("examples/landscape.yaml", "landscape")
    config.run()


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples not present - skipping examples tests")
def test_finance_e2e():
    config = Configuration.load("examples/finance.yaml", "finance")
    config.run()


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples not present - skipping examples tests")
def test_research_e2e():
    config = Configuration.load("examples/research.yaml", "research")
    config.run()


@pytest.mark.skipif(_example_folder_does_not_exist(), reason="Examples not present - skipping examples tests")
@pytest.mark.parametrize("parameters", ("string1", "string2"))
def test_hydra_e2e(parameters):
    run_hydra("examples/hydra.yaml", ["config=inline", "page=report", f"parameters={parameters}"])



from IPython.display import HTML

from nbprint import Content, ContentFlexColumnLayout, ContentFlexRowLayout, Role

__all__ = (
    "ExampleResearchHeader",
    "ExampleResearchTitle",
    "ExampleResearchAuthor",
    "ExampleResearchAuthors",
    "ExampleResearchBody",
    "ExampleResearchSectionTitle",
    "ExampleResearchSectionText",
)


class ExampleResearchHeader(ContentFlexColumnLayout):
    role: Role = Role.LAYOUT
    css: str = """
:scope {
  padding-top: 50px;
  display: flex;
  flex-direction: column;
  padding-bottom: 50px;
}
"""

    def __call__(self, ctx=None, *args, **kwargs):
        # return empty html just for placeholder
        return HTML("")


class ExampleResearchTitle(Content):
    title: str

    css: str = """
:scope {
}

h1 {
  font-family: serif;
  font-weight: bold !important;
  font-size: 25px !important;
  margin: auto !important;
}
    """

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"""
        <div class="row">
            <h1>{self.title}</h1>
        </div>
        """)


class ExampleResearchAuthor(Content):
    name: str
    email: str

    css: str = """
:scope {
  text-align: center;
}
div.column > span:nth-child(1) {
  font-size: 16px;
  font-family: serif !important;
}
div.column > span:nth-child(2) {
  font-size: 12px;
  font-family: sans-serif !important;
}
    """

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"""
        <div class="column">
            <span>{self.name}</span>
            <span>{self.email}</span>
        </div>
        """)


class ExampleResearchAuthors(ContentFlexRowLayout): ...


class ExampleResearchBody(Content):
    role: Role = Role.LAYOUT
    css: str = """
:scope {
  column-count: 2;
  -webkit-column-count: 2;
  column-gap: 1cm;
}
"""

    def __call__(self, ctx=None, *args, **kwargs):
        # return empty html just for placeholder
        return HTML("")


class ExampleResearchSectionTitle(Content):
    title: str

    css: str = """
  h1 {
    padding-bottom: 10px;
  }
  """

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"<h1>{self.title}</h1>")


class ExampleResearchSectionText(Content):
    text: str

    def __call__(self, ctx=None, *args, **kwargs):
        return HTML(f"<p>{self.text}</p>")



import os
from datetime import date, datetime
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Optional
from uuid import uuid4

from jinja2 import Template
from nbformat import NotebookNode, writes
from pydantic import DirectoryPath, Field, field_validator

from nbprint.config.base import BaseModel, Role

if TYPE_CHECKING:
    from .config import Configuration

__all__ = ("Outputs", "NBConvertOutputs")


class Outputs(BaseModel):
    path_root: DirectoryPath = Field(default=Path.cwd())
    naming: str = Field(default="{{name}}-{{date}}")

    tags: list[str] = Field(default=["nbprint:outputs"])
    role: Role = Role.OUTPUTS
    ignore: bool = True

    @field_validator("path_root", mode="before")
    @classmethod
    def convert_str_to_path(cls, v) -> Path:
        if isinstance(v, str):
            v = Path(v)
        if isinstance(v, Path):
            v.resolve().mkdir(parents=True, exist_ok=True)
            return v
        raise TypeError

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.path_root = self.path_root.resolve()
        self.path_root.mkdir(parents=True, exist_ok=True)

    def _get_name(self, config: "Configuration") -> str:
        return config.name

    def _get_date(self, **_) -> str:
        return date.today().isoformat()

    def _get_datetime(self, **_) -> str:
        return datetime.now().isoformat()

    def _get_uuid(self, **_) -> str:
        return uuid4()

    def _get_sha(self, config: "Configuration") -> str:
        import hashlib

        return hashlib.sha256(config.model_dump_json(by_alias=True).encode()).hexdigest()

    def run(self, config: "Configuration", gen: NotebookNode) -> Path:
        # create file or folder path
        name = Template(self.naming).render(
            name=self._get_name(config=config),
            date=self._get_date(config=config),
            datetime=self._get_datetime(config=config),
            uuid=self._get_uuid(config=config),
            sha=self._get_sha(config=config),
        )
        file = Path(str(Path(self.path_root).resolve() / f"{name}.ipynb"))
        file.write_text(writes(gen))
        return file

    def generate(self, metadata: dict, config: "Configuration", parent: BaseModel, **kwargs) -> NotebookNode:
        return super().generate(metadata=metadata, config=config, parent=parent, attr="outputs", **kwargs)


class NBConvertOutputs(Outputs):
    target: Optional[Literal["ipynb", "html", "pdf", "webpdf"]] = "html"  # TODO: nbconvert types
    execute: Optional[bool] = True
    timeout: Optional[int] = 600
    template: Optional[str] = "nbprint"

    @field_validator("target", mode="before")
    @classmethod
    def validate_target(cls, v) -> str:
        if v is None:
            return "html"
        if v == "pdf":
            return "webpdf"
        return v

    def run(self, config: "Configuration", gen: NotebookNode) -> Path:
        from nbconvert.nbconvertapp import main as execute_nbconvert

        # set for convenience
        os.environ["PSP_JUPYTER_HTML_EXPORT"] = "1"

        # run the nbconvert
        notebook = super().run(config=config, gen=gen)

        cmd = [
            str(notebook),
            f"--to={self.target}",
            f"--template={self.template}",
        ]

        if self.execute:
            cmd.extend(
                [
                    "--execute",
                    f"--ExecutePreprocessor.timeout={self.timeout}",
                ]
            )

        execute_nbconvert(cmd)
        return Path(str(notebook).replace(".ipynb", f".{self.target}"))


# class PapermillOutputs(NBConvertOutputs):
#     def run(self, config: "Configuration", gen: NotebookNode) -> Path:
#         from nbconvert.nbconvertapp import main

#         notebook = super().run(config=config, gen=gen)
#         main([notebook, f"--to={self.target}", f"--template={self.template}", "--execute"])



from pathlib import Path
from pprint import pprint
from sys import version_info
from typing import Optional, Union

from hydra import compose, initialize_config_dir
from hydra.utils import instantiate
from nbformat import NotebookNode
from nbformat.v4 import new_notebook
from pydantic import Field, PrivateAttr, field_validator

from nbprint import __version__
from nbprint.config.base import BaseModel, Role, Type, _append_or_extend
from nbprint.config.content import Content
from nbprint.config.exceptions import NBPrintPathOrModelMalformedError
from nbprint.config.page import Page

from .context import Context
from .outputs import Outputs
from .parameters import Parameters

__all__ = (
    "Configuration",
    "load",
)


class Configuration(BaseModel):
    name: str
    resources: dict[str, BaseModel] = Field(default_factory=dict)
    outputs: Outputs
    parameters: Parameters = Field(default_factory=Parameters)
    page: Page = Field(default_factory=Page)
    context: Context = Field(default_factory=Context)
    content: list[Content] = Field(default_factory=list)

    # basic metadata
    tags: list[str] = Field(default=["nbprint:config"])
    role: Role = Role.CONFIGURATION
    ignore: bool = True
    debug: bool = True

    # internals
    _nb_var_name: str = PrivateAttr(default="nbprint_config")
    _nb_vars: set = PrivateAttr(default_factory=set)

    @field_validator("resources", mode="before")
    @classmethod
    def convert_resources_from_obj(cls, value) -> dict[str, BaseModel]:
        if value is None:
            value = {}
        if isinstance(value, dict):
            for k, v in value.items():
                value[k] = BaseModel._to_type(v)
        return value

    @field_validator("outputs", mode="before")
    @classmethod
    def convert_outputs_from_obj(cls, v) -> Outputs:
        return BaseModel._to_type(v, Outputs)

    @field_validator("parameters", mode="before")
    @classmethod
    def convert_parameters_from_obj(cls, v) -> Parameters:
        return BaseModel._to_type(v, Parameters)

    @field_validator("page", mode="before")
    @classmethod
    def convert_page_from_obj(cls, v) -> Page:
        return BaseModel._to_type(v, Page)

    @field_validator("context", mode="before")
    @classmethod
    def convert_context_from_obj(cls, v) -> Context:
        return BaseModel._to_type(v, Context)

    @field_validator("content", mode="before")
    @classmethod
    def convert_content_from_obj(cls, v) -> Content:
        if v is None:
            return []
        if isinstance(v, list):
            for i, element in enumerate(v):
                if isinstance(element, str):
                    v[i] = Content(type=Type.from_string(element))
                elif isinstance(element, dict):
                    v[i] = BaseModel._to_type(element)
        return v

    def generate(self, **_) -> list[NotebookNode]:
        nb = new_notebook()
        nb.metadata.nbprint = {}
        nb.metadata.nbprint.version = __version__
        nb.metadata.nbprint.tags = []
        nb.metadata.nbprint.nbprint = {}
        nb.metadata.nbprint.language = f"python{version_info.major}.{version_info.minor}"

        base_meta = {
            "tags": [],
            "nbprint": {},
        }

        nb.cells = []

        # start with parameters for papermill compat
        # use `parent=None` because we parameters is first cell, and we wont instantiate the config
        # until the next cell
        _append_or_extend(nb.cells, self.parameters.generate(metadata=base_meta.copy(), config=self, parent=None))

        # now do the configuration itself
        _append_or_extend(nb.cells, self._generate_self(metadata=base_meta.copy()))

        # now do the context object
        # pass in parent=self, attr=context so we do config.context
        _append_or_extend(nb.cells, self.context.generate(metadata=base_meta.copy(), config=self, parent=self, attr="context"))

        # resources: dict[str, SerializeAsAny[BaseModel]] = Field(default_factory=dict)
        # TODO: omitting resources, referenced directly in yaml
        # cell.metadata.nbprint.resources = {k: v.model_dump_json(by_alias=True) for k, v in self.resources.items()}

        # outputs: SerializeAsAny[Outputs]
        # TODO: skipping, consumed internally

        # now setup the page layout
        # pass in parent=self, attr=page so we do config.page
        _append_or_extend(nb.cells, self.page.generate(metadata=base_meta.copy(), config=self, parent=self, attr="page"))

        # now iterate through the content, recursively generating
        for i, content in enumerate(self.content):
            _append_or_extend(
                nb.cells,
                content.generate(metadata=base_meta.copy(), config=self, parent=self, attr="content", counter=i),
            )

        return nb

    def _generate_self(self, metadata: dict) -> NotebookNode:
        cell = super()._base_generate(metadata=metadata, config=self)

        # omit the data
        cell.metadata.nbprint.data = ""

        # add extras
        cell.metadata.nbprint.debug = self.debug

        # add resources
        # TODO: do this or no?
        # cell.metadata.nbprint.resources = {k: v.model_dump_json(by_alias=True) for k, v in self.resources.items()}
        cell.metadata.nbprint.outputs = self.outputs.model_dump_json(by_alias=True)
        return cell

    def _generate_resources_cells(self, metadata: Optional[dict] = None) -> NotebookNode:
        cell = super()._base_generate(metadata=metadata, config=None)

        # omit the data
        cell.metadata.nbprint.data = ""

        # add resources
        # mod = ast.Module(body=[], type_ignores=[])
        # for k, v in self.resources.items():
        #     ...
        return cell

    @staticmethod
    def load(path_or_model: Union[str, Path, dict, "Configuration"], name: str) -> "Configuration":
        if isinstance(path_or_model, Configuration):
            return path_or_model

        if isinstance(path_or_model, str) and path_or_model.endswith((".yml", ".yaml")):
            path_or_model = Path(path_or_model).resolve()

        if isinstance(path_or_model, Path):
            path_or_model = path_or_model.resolve()
            folder = str(path_or_model.parent)
            file = str(path_or_model.name)

            with initialize_config_dir(version_base=None, config_dir=folder, job_name=name):
                cfg = compose(config_name=file, overrides=[f"+name={name}"])
                config = instantiate(cfg)
                if not isinstance(config, Configuration):
                    config = Configuration(**config)
                return config
        raise NBPrintPathOrModelMalformedError(path_or_model)

    def run(self, dry_run: bool = False) -> None:
        gen = self.generate()
        if self.debug:
            pprint(gen)
        if not dry_run:
            self.outputs.run(self, gen)

        # reset in case we want to run again
        self._reset()

    def _reset(self) -> None:
        # reset ourselves in case we need to rerun
        self._nb_vars = set()
        self.context._context_generated = False


load = Configuration.load



from pathlib import Path
from typing import Optional

from IPython.display import HTML, Image
from pydantic import Field, FilePath, field_validator

from .base import Content


class ContentImage(Content):
    path: Optional[FilePath] = None
    content: Optional[bytes] = b""
    tags: list[str] = Field(default=["nbprint:content", "nbprint:content:image"])

    @field_validator("path", mode="before")
    @classmethod
    def convert_path_from_obj(cls, v) -> Path:
        if isinstance(v, str):
            v = Path(v).resolve()
        return v

    @field_validator("content", mode="before")
    @classmethod
    def convert_content_from_obj(cls, v) -> bytes:
        if v is None:
            return b""
        if v and isinstance(v, str):
            v = Path(v).resolve()
        if v and isinstance(v, Path):
            v = v.read_bytes()
        return v

    def as_base64(self) -> str:
        img = Image(filename=self.path) if self.path else Image(data=self.content)
        return img._repr_png_()

    def __call__(self, **_) -> HTML:
        return HTML(f"""<img src="data:image/png;base64,{self.as_base64()}">""")

    def __repr__(self) -> str:
        return f"Image(path='{self.path}', content=[{len(self.content)} bytes])"



__all__ = (
    "NBPrintGenerationError",
    "NBPrintConfigurationError",
    "NBPrintNullCellError",
    "NBPrintBadScopeError",
    "NBPrintPathOrModelMalformedError",
    "NBPrintGenerationError",
    "NBPrintConfigurationError",
)


class NBPrintGenerationError(RuntimeError): ...


class NBPrintConfigurationError(RuntimeError): ...


class NBPrintNullCellError(NBPrintGenerationError):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__("got null cell, investigate!", *args, **kwargs)


class NBPrintBadScopeError(NBPrintConfigurationError):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__("Must set one of element, classname, id, or selector", *args, **kwargs)


class NBPrintPathOrModelMalformedError(NBPrintConfigurationError):
    def __init__(self, path_or_model, *args, **kwargs) -> None:
        super().__init__(f"Path or model malformed: {path_or_model} {type(path_or_model)}", *args, **kwargs)



from .base import Content, ContentCode, ContentMarkdown
from .common import TextComponent
from .cover import ContentCover
from .image import ContentImage
from .page import *
from .pagebreak import ContentPageBreak
from .table_of_contents import ContentTableOfContents



from pathlib import Path

import pytest

from nbprint.cli import run_hydra


@pytest.mark.parametrize("file", list((Path(__file__).parent / "files").glob("*")))
def test_integrations(file):
    run_hydra(str(file.resolve()))



from nbprint import (
    Border,
    BorderLineStyle,
    BorderStyle,
    Element,
    Font,
    FontFamily,
    FontStyle,
    FontWeight,
    Padding,
    Scope,
    Spacing,
    Style,
    TextTransform,
)


class TestStyles:
    def test_basic_style_scope(self):
        s = Style(
            scope=Scope(
                element="div",
                classname="row",
            ),
            spacing=Spacing(padding=Padding(bottom=10)),
        )

        assert str(s) == "div.row {\npadding-bottom: 10px;\n}"

    def test_basic_style_text_padding_border(self):
        s = Style(
            scope=Scope(
                element="h1",
            ),
            font=Font(
                family=FontFamily.sans_serif,
                size=25,
            ),
            spacing=Spacing(padding=Padding(right=15)),
            border=Border(
                right=BorderStyle(
                    width=2,
                    style=BorderLineStyle.solid,
                    color="black",
                )
            ),
        )
        assert str(s) == "h1 {\npadding-right: 15px;\nfont-family: sans-serif;\nfont-size: 25px;\nborder-right: 2px solid black;\n}"

    def test_basic_style_text_padding_border2(self):
        s = Style(
            scope=Scope(
                element=Element.h2,
            ),
            font=Font(
                family=FontFamily.sans_serif,
                size=12,
                transform=TextTransform.uppercase,
                weight=FontWeight.bold,
            ),
            spacing=Spacing(padding=Padding(left=15)),
        )

        assert str(s) == "h2 {\npadding-left: 15px;\nfont-family: sans-serif;\nfont-size: 12px;\ntext-transform: uppercase;\nfont-weight: bold;\n}"

    def test_basic_style_text_padding_border3(self):
        s = Style(
            scope=Scope(
                element=Element.span,
            ),
            font=Font(size=15, style=FontStyle.italic, color="grey"),
        )

        assert str(s) == "span {\nfont-size: 15px;\nfont-style: italic;\ncolor: grey;\n}"



import os.path

from IPython.display import HTML, Image
from pydantic import Field, FilePath

from nbprint import Configuration, ContentImage, Page, PageOrientation, PageRegion, PageRegionContent, Role

from .cover import ExampleNBPrintLogo

__all__ = (
    "ExampleLogoInFooter",
    "ExampleBookIcon",
    "ExampleReportPage",
)


class ExampleLogoInFooter(PageRegion):
    logo: ContentImage = Field(default_factory=ExampleNBPrintLogo)
    content: PageRegionContent = Field(default_factory=lambda: PageRegionContent(content="element(footerLogo)"))

    ignore: bool = False

    css: str = """
.footer-logo {
  position: running(footerLogo);
  width: 75px;
}
"""

    def __call__(self, *args, **kwargs):
        img = Image(filename=self.logo.path) if self.logo.path else Image(data=self.logo.content)
        return HTML(f"""<img class="footer-logo" src="data:image/png;base64,{img._repr_png_()}">""")


class ExampleBookIcon(ContentImage):
    path: FilePath = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "..", "examples", "assets", "book.png"))


class ExampleReportPage(Page):
    bottom_left: PageRegion = Field(default=ExampleLogoInFooter())
    bottom: PageRegion = Field(default=PageRegion(content=PageRegionContent(content="Example Report")))
    bottom_right: PageRegion = Field(default=PageRegion())
    role: Role = Role.PAGE

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.css += f"""
.pagedjs_first_page {{
    background-image: url("data:image/png;base64,{ExampleBookIcon().as_base64()}");
    background-size: 50%;
    background-repeat: no-repeat;
}}

.pagedjs_first_page div.pagedjs_margin-bottom {{
    display: none;
}}

.pagedjs_margin-bottom-center > pagedjs_margin-content::after {{
    font-weight: bold;
}}
"""

    def render(self, config: Configuration) -> None:
        super().render(config=config)
        if config.page.orientation == PageOrientation.landscape:
            self.css += """
.pagedjs_first_page {
    background-position-y: 85%;
    background-position-x: 95%;
}
"""
        else:
            self.css += """
.pagedjs_first_page {
    background-position-y: 75%;
    background-position-x: 50%;
}
"""



from pathlib import Path

from nbprint.cli import run_hydra


def test_config_rerunsafe():
    config = run_hydra(str(Path(__file__).parent / "files" / "hermetic.yaml"), dry_run=True)
    config1 = config.generate()
    config._reset()
    config2 = config.generate()
    code1 = ""
    code2 = ""
    for cell in config1.cells:
        code1 += cell["source"] + "\n"
    for cell in config2.cells:
        code2 += cell["source"] + "\n"
    assert code1 == code2



from nbprint import Content

__all__ = ("ExampleCodeBlock",)


class ExampleCodeBlock(Content):
    def __call__(self, ctx=None, *args, **kwargs):
        from great_tables import GT

        return (
            GT(ctx.df)
            .tab_header(title="Sample Table")
            .tab_source_note("This is where authors provide additional information about the data, including whatever notes are needed.")
            .fmt_number(decimals=2)
        )



from typing import Optional, Union

from strenum import StrEnum

from .common import Color
from .css import _BaseCss


class FontWeight(StrEnum):
    normal = "normal"
    bold = "bold"


class FontStyle(StrEnum):
    normal = "normal"
    italic = "italic"


class TextDecoration(StrEnum):
    none = "none"
    underline = "underline"
    overline = "overline"
    dotted = "dotted"
    wavy = "wavy"


class TextTransform(StrEnum):
    capitalize = "capitalize"
    lowercase = "lowercase"
    uppercase = "uppercase"


class FontFamily(StrEnum):
    serif = "serif"
    sans_serif = "sans-serif"
    monospace = "monospace"


class Font(_BaseCss):
    family: Optional[Union[FontFamily, str]] = None
    size: Optional[int] = None
    transform: Optional[TextTransform] = None

    # TODO: multiple decorations and color decorations
    # text-decoration-color
    # text-decoration-line
    # text-decoration-style
    # text-decoration-thickness

    decoration: Optional[TextDecoration] = None

    style: Optional[FontStyle] = None
    weight: Optional[FontWeight] = None
    color: Optional[Color] = None

    def __str__(self) -> str:
        ret = ""
        if self.family:
            ret += f"font-family: {self.family};\n"
        if self.size:
            ret += f"font-size: {self.size}px;\n"
        if self.transform:
            ret += f"text-transform: {self.transform};\n"
        if self.decoration:
            ret += f"text-decoration: {self.decoration};\n"
        if self.style:
            ret += f"font-style: {self.style};\n"
        if self.weight:
            ret += f"font-weight: {self.weight};\n"
        if self.color:
            ret += f"color: {self.color};\n"
        return ret.strip()


class Text(Font): ...



from IPython.display import HTML, display
from pydantic import Field

from .base import Content


class ContentTableOfContents(Content):
    tags: list[str] = Field(default=["nbprint:content", "nbprint:content:table-of-contents"])

    def __call__(self, **_) -> None:
        display(HTML('<div id="toc"></div>'))
        # display(HTML('<p class="pagebreak"></p>'))



from typing import Optional

from IPython.display import Markdown
from pydantic import Field

from nbprint.config.common import FontStyle, FontWeight, HorizontalAlignment, TextDecoration, VerticalAlignment

from .base import Content


class TextComponent(Content):
    text: Optional[str] = ""
    horizontal_alignment: Optional[HorizontalAlignment] = None
    vertical_alignment: Optional[VerticalAlignment] = None
    font_weight: Optional[FontWeight] = None
    font_style: Optional[FontStyle] = None
    text_decoration: Optional[TextDecoration] = None

    tags: list[str] = Field(default=["nbprint:content"])

    def __call__(self, **_) -> Markdown:
        return Markdown(self.text)



from typing import TYPE_CHECKING, Optional, Union

from IPython.display import HTML
from nbformat import NotebookNode
from pydantic import Field, field_validator

from nbprint.config.base import BaseModel, Role, SerializeAsAny, Type, _append_or_extend
from nbprint.config.common import Style
from nbprint.config.exceptions import NBPrintNullCellError

if TYPE_CHECKING:
    from nbprint.config.core import Configuration


class Content(BaseModel):
    content: Optional[Union[str, list[SerializeAsAny[BaseModel]]]] = ""
    tags: list[str] = Field(default=["nbprint:content"])
    role: Role = Role.CONTENT

    # used by lots of things
    style: Optional[Style] = None

    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["Configuration"] = None,
        parent: Optional["BaseModel"] = None,
        attr: str = "",
        counter: Optional[int] = None,
    ) -> Optional[Union[NotebookNode, list[NotebookNode]]]:
        # make a cell for yourself
        self_cell = super().generate(
            metadata=metadata,
            config=config,
            parent=parent,
            attr=attr or "content",
            counter=counter,
        )
        if isinstance(self.content, str) and self.content:
            # replace content if its a str
            self_cell.source = self.content

            # remove the data, redundant
            self_cell.metadata.nbprint.data = ""

        cells = [self_cell]

        if isinstance(self.content, list):
            for i, cell in enumerate(self.content):
                _append_or_extend(
                    cells,
                    cell.generate(metadata=metadata, config=config, parent=self, attr=attr or "content", counter=i),
                )
        for cell in cells:
            if cell is None:
                raise NBPrintNullCellError
        return cells

    @field_validator("content", mode="before")
    @classmethod
    def convert_content_from_obj(cls, v) -> "Content":
        if v is None:
            return []
        if isinstance(v, list):
            for i, element in enumerate(v):
                if isinstance(element, str):
                    v[i] = Content(type=Type.from_string(element))
                elif isinstance(element, dict):
                    v[i] = BaseModel._to_type(element)
        return v

    def __call__(self, **_) -> HTML:
        return HTML("")


class ContentMarkdown(Content):
    content: Optional[str] = ""

    def generate(
        self,
        metadata: Optional[dict] = None,
        config: Optional["Configuration"] = None,
        parent: Optional["BaseModel"] = None,
        **_,
    ) -> Optional[Union[NotebookNode, list[NotebookNode]]]:
        cell = super()._base_generate_md(metadata=metadata, config=config, parent=parent)
        cell.source = self.content
        return cell


class ContentCode(Content): ...



from typing import Literal, Optional

from pydantic_extra_types.color import Color
from strenum import StrEnum

from .css import _BaseCss

__all__ = (
    "Color",  # reexport
    "Unset",
    "Element",
    "Direction",
    "PseudoClass",
    "PseudoElement",
)

Unset = Literal["unset"]

# class Size:
# cm Centimeters 1cm = 37.8px = 25.2/64in
# mm Millimeters 1mm = 1/10th of 1cm
# Q Quarter-millimeters 1Q = 1/40th of 1cm
# in Inches 1in = 2.54cm = 96px
# pc Picas 1pc = 1/6th of 1in
# pt Points 1pt = 1/72nd of 1in
# px Pixels 1px = 1/96th of 1in
# em Font size of the parent, in the case of typographical properties like font-size, and font size of the element itself,
#     in the case of other properties like width.
# ex x-height of the element's font.
# ch The advance measure (width) of the glyph "0" of the element's font.
# rem Font size of the root element.
# lh Line height of the element.
# rlh Line height of the root element. When used on the font-size or line-height properties of the root element, it refers
#     to the properties' initial value.
# vw 1% of the viewport's width.
# vh 1% of the viewport's height.
# vmin 1% of the viewport's smaller dimension.
# vmax 1% of the viewport's larger dimension.
# vb 1% of the size of the initial containing block in the direction of the root element's block axis.
# vi 1% of the size of the initial containing block in the direction of the root element's inline axis.
# svw, svh 1% of the small viewport's width and height, respectively.
# lvw, lvh 1% of the large viewport's width and height, respectively.
# dvw, dvh 1% of the dynamic viewport's width and height, respectively.


class Element(StrEnum):
    div = "div"
    span = "span"
    p = "p"
    h1 = "h1"
    h2 = "h2"
    h3 = "h3"
    h4 = "h4"
    h5 = "h5"
    h6 = "h6"


class PseudoClass(StrEnum):
    active = "active"
    checked = "checked"
    disabled = "disabled"
    empty = "empty"
    enabled = "enabled"
    first_child = "first-child"
    # :first-of-type
    focus = "focus"
    hover = "hover"
    # :in-range
    # :invalid
    # :lang(language) p:lang(it) Selects every <p> element with a lang attribute value starting with "it"
    last_child = "last-child"
    # :last-of-type
    # :link
    # :not(selector) :not(p) Selects every element that is not a <p> element
    # :nth-child(n) p:nth-child(2) Selects every <p> element that is the second child of its parent
    # :nth-last-child(n) p:nth-last-child(2) Selects every <p> element that is the second child of its parent, counting from the last child
    # :nth-last-of-type(n) p:nth-last-of-type(2) Selects every <p> element that is the second <p> element of its parent, counting from the last child
    # :nth-of-type(n) p:nth-of-type(2) Selects every <p> element that is the second <p> element of its parent
    # :only-of-type
    # :only-child
    # :optional
    # :out-of-range
    # :read-only
    # :read-write
    # :required
    # :root
    # :target
    # :valid
    visited = "visited"


class PseudoElement(StrEnum):
    after = "after"
    # backdrop = "backdrop"
    before = "before"
    # cue = "cue"
    # cue_region = "cue-region"
    first_letter = "first-letter"
    first_line = "first-line"
    # file_selector_button = "file-selector-button"
    # grammar_error = "grammar-error"
    marker = "marker"
    # part = "part(..."
    placeholder = "placeholder"
    selection = "selection"
    # slotted = "slotted(..."
    # spelling_error = "spelling-error"
    # target_text = "target-text"


class Direction(StrEnum):
    left = "left"
    right = "right"
    top = "top"
    bottom = "bottom"


class _DirectionalSize(_BaseCss):
    right: Optional[int] = None
    left: Optional[int] = None
    top: Optional[int] = None
    bottom: Optional[int] = None



from typing import TYPE_CHECKING

from nbformat import NotebookNode
from pydantic import Field, PrivateAttr

from nbprint.config.base import BaseModel, Role

if TYPE_CHECKING:
    from .config import Configuration

__all__ = ("Context",)


class Context(BaseModel):
    tags: list[str] = Field(default=["nbprint:context"])
    role: Role = Role.CONTEXT
    ignore: bool = True

    # internals
    _nb_var_name: str = PrivateAttr(default="nbprint_ctx")
    _context_generated: bool = PrivateAttr(default=False)

    class Config:
        """Pydantic configuration object."""

        arbitrary_types_allowed: bool = True
        extra: str = "allow"
        validate_assignment: bool = False

    def generate(self, metadata: dict, config: "Configuration", parent: BaseModel, attr: str = "", **kwargs) -> NotebookNode:
        self._context_generated = True
        return super().generate(metadata=metadata, config=config, parent=parent, attr=attr, **kwargs)



from pathlib import Path
from typing import Optional

from hydra import compose, initialize_config_dir
from hydra.utils import instantiate
from typer import Argument, Typer

from .config import Configuration


def run(path: Path, name: str, dry_run: bool = False) -> Configuration:
    config = Configuration.load(path, name)
    config.run(dry_run=dry_run)
    return config


def run_hydra(path: str = "", overrides: Optional[list[str]] = Argument(None), dry_run: bool = False) -> Configuration:  # noqa: B008
    path = Path(path)
    if not isinstance(overrides, list):
        # maybe running via python, reset
        overrides = []
    with initialize_config_dir(config_dir=str(path.parent.absolute()), version_base=None):
        cfg = compose(config_name=str(path.name), overrides=overrides)
        config = instantiate(cfg)
        if not isinstance(config, Configuration):
            config = Configuration(**config)
        config.run(dry_run=dry_run)
    return config


def main() -> None:
    app = Typer()
    app.command("run")(run)
    app.command("hydra")(run_hydra)
    app()



from typing import Optional, Union

from strenum import StrEnum

from .common import Color
from .css import _BaseCss

__all__ = (
    "BorderLineStyle",
    "BorderLineWidth",
    "BorderStyle",
    "Border",
)


class BorderLineStyle(StrEnum):
    none = "none"
    hidden = "hidden"
    dotted = "dotted"
    dashed = "dashed"
    solid = "solid"
    double = "double"
    groove = "groove"
    ridge = "ridge"
    inset = "inset"
    outset = "outset"


class BorderLineWidth(StrEnum):
    thin = "thin"
    medium = "medium"
    thick = "thick"


class BorderStyle(_BaseCss):
    width: Union[BorderLineWidth, int]
    style: BorderLineStyle
    color: Color

    def __str__(self) -> str:
        return f"{self.width}px {self.style} {self.color}"


class Border(_BaseCss):
    right: Optional[BorderStyle] = None
    left: Optional[BorderStyle] = None
    top: Optional[BorderStyle] = None
    bottom: Optional[BorderStyle] = None

    def __str__(self) -> str:
        return "\n".join(
            f"border-{direction}: {getattr(self, direction, '')};" for direction in ("right", "left", "top", "bottom") if getattr(self, direction, "")
        ).strip()



from strenum import StrEnum

__all__ = (
    "PageSize",
    "PageOrientation",
)


class PageSize(StrEnum):
    A0 = "A0"  # 841 x 1189 mm
    A1 = "A1"  # 594 x 841 mm
    A2 = "A2"  # 420 x 594 mm
    A3 = "A3"  # 297 x 420 mm
    A4 = "A4"  # 210 x 297 mm
    A5 = "A5"  # 148 x 210 mm
    A6 = "A6"  # 105 x 148 mm
    A7 = "A7"  # 74 x 105 mm
    A10 = "A10"  # 26 x 37 mm
    B4 = "B4"  # 250 x 353 mm
    B5 = "B5"  # 176 x 250 mm
    letter = "letter"  # 8.5 x 11 in
    legal = "legal"  # 8.5 x 14 in
    ledger = "ledger"  # 11 x 17 in


class PageOrientation(StrEnum):
    portrait = "portrait"
    landscape = "landscape"



from typing import Optional, Union

from pydantic import model_validator

from nbprint.config.base import BaseModel
from nbprint.config.exceptions import NBPrintBadScopeError

from .border import Border
from .common import Element, PseudoClass, PseudoElement
from .spacing import Spacing
from .text import Font


class Scope(BaseModel):
    element: Optional[Union[Element, str]] = ""

    id: Optional[str] = ""
    classname: Optional[str] = ""

    selector: Optional[str] = ""

    pseudoclass: Optional[PseudoClass] = ""
    pseudoelement: Optional[PseudoElement] = ""

    @model_validator(mode="after")
    def check_any_set(self) -> "Scope":
        if all(element in ("", None) for element in (self.element, self.classname, self.id, self.selector)):
            raise NBPrintBadScopeError
        return self

    def __str__(self) -> str:
        ret = ""
        if self.element:
            ret = f"{self.element}"
        if self.id:
            ret = f"{ret}#{self.id}"
        if self.classname:
            ret = f"{ret}.{self.classname}"
        if self.selector:
            ret = f"{ret}[{self.selector}]"
        if self.pseudoclass:
            ret = f"{ret}:{self.pseudoclass}"
        if self.pseudoelement:
            ret = f"{ret}::{self.pseudoclass}"
        return ret


class Style(BaseModel):
    scope: Optional[Scope] = None
    spacing: Optional[Spacing] = None
    font: Optional[Font] = None
    border: Optional[Border] = None

    def __str__(self) -> str:
        ret = f"{self.scope} {{\n"
        for part in (self.spacing, self.font, self.border):
            if part:
                part_str = str(part)
                if part_str:
                    ret += f"{part_str}\n"
        ret += "\n}"
        return ret.replace("\n\n", "\n").strip()
